#include <R.h>
#include <Rmath.h>
#include <R_ext/BLAS.h>
#include <R_ext/Lapack.h>
#include <R_ext/Utils.h>

void sqmatmul(double *A, double *B, int *n, double *res) {
	double sum;
	
	for(int i=0; i<*n; i++) {
		for(int j=0; j<*n; j++) {
			sum = 0.0;
			for(int k=0; k<*n; k++) {
				sum = sum + A[i+(*n)*k] * B[k+(*n)*j];
			}
			res[i+(*n)*j] = sum;
		}
	}
}

/* Example of sqmatmul

> A = matrix(rnorm(9), 3)
> B = matrix(rnorm(9), 3)
> x = .C("sqmatmul", as.double(A), as.double(B), as.integer(3), res=as.double(matrix(0, nrow=3, ncol=3)))
> x$res
[1]  0.8598526 -0.5747655  0.5686713 -1.1030245  2.6063841 -2.3720913
[7]  0.9367630 -1.8974657  1.6676321
> matrix(x$res, 3)
           [,1]      [,2]      [,3]
[1,]  0.8598526 -1.103025  0.936763
[2,] -0.5747655  2.606384 -1.897466
[3,]  0.5686713 -2.372091  1.667632
> A %*% B
           [,1]      [,2]      [,3]
[1,]  0.8598526 -1.103025  0.936763
[2,] -0.5747655  2.606384 -1.897466
[3,]  0.5686713 -2.372091  1.667632

*/

// ====================================================== \\



void rw2d(int *n, int *xRes, int *yRes) {
	GetRNGstate();
	
	int *xRes_p, *yRes_p, lastx, lasty;
	xRes_p = xRes;
	yRes_p = yRes;
	lastx = lasty = *(xRes_p++) = *(yRes_p++) = 0;
	
	for(int i=1; i<*n; i++) {
		if(runif(0.0, 1.0) < 0.5) {
			lastx = *(xRes_p++) = lastx + sign(runif(-1.0, 1.0));
			*(yRes_p++) = lasty;
		} else {
			*(xRes_p++) = lastx;
			lasty = *(yRes_p++) = lasty + sign(runif(-1.0, 1.0));
		}
	}
	
	PutRNGstate();
}

/* Example of rw2d

> walk = .C("rw2d", as.integer(20), x=as.integer(vector("numeric", 20)), y=as.integer(vector("numeric", 20)))
> walk$x
 [1]  0  0 -1 -1 -2 -1  0  0  0 -1 -1 -1 -2 -3 -4 -5 -5 -4 -5 -6
> walk$y
 [1]  0 -1 -1  0  0  0  0 -1  0  0 -1  0  0  0  0  0  1  1  1  1
> walk = .C("rw2d", as.integer(10000), x=as.integer(vector("numeric", 10000)), y=as.integer(vector("numeric", 10000)))
> plot(walk$x, walk$y, type="l", col="#0000CC55")

*/

// ====================================================== \\










void linmod(double *X, double *y, int *n, int *p, double *res) {
	double *XtX, *Xty;
	XtX = (double *) R_alloc(*p * *p, sizeof(double));
	Xty = (double *) R_alloc(*p, sizeof(double));
	
	// X^t X
	char transa = 'T', transb = 'N';
	double alpha = 1.0, beta = 0.0;
	F77_CALL(dgemm)(&transa, &transb, p, p, n, &alpha, X, n, X, n, &beta, XtX, p);
	
	// X^t y
	int inc = 1;
	F77_CALL(dgemv)(&transa, n, p, &alpha, X, n, y, &inc, &beta, Xty, &inc);
	
	// (X^t X)^{-1}
	// N.B: LAPACK requires the LU decomposition first, then can do the inverse
	// So, LU decompose ...
	int *ipiv, info;
	ipiv = (int *) R_alloc(*p, sizeof(int));
	F77_CALL(dgetrf)(p, p, XtX, p, ipiv, &info);
	// ... then invert
	// 1st get the optimal workspace size from LAPACK
	double *work;
	work = (double *) R_alloc(1, sizeof(double));
	int lwork = -1;
	F77_CALL(dgetri)(p, XtX, p, ipiv, work, &lwork, &info);
	// 2nd assign this workspace and do the inverse
	lwork = (int) *work;
	work = (double *) R_alloc(lwork, sizeof(double));
	F77_CALL(dgetri)(p, XtX, p, ipiv, work, &lwork, &info);
	
	// Finally: (X^t X)^{-1} X^t y
	F77_CALL(dgemv)(&transb, p, p, &alpha, XtX, p, Xty, &inc, &beta, res, &inc);
}

/* Example of linmod

> # Generate
> X = matrix(runif(1000,-5,5), nrow=100, ncol=10)
> beta = rexp(10,2)
> intercept = rexp(1,4)
> y = intercept + X %*% beta + rnorm(100)
> 
> # Fit
> reslm = lm(y ~ X)
> res = .C("linmod", c(rep(1,100), as.double(X)), as.double(y), as.integer(100), as.integer(11), coef=as.double(1:11))
> 
> # True -vs- fit
> c(intercept, beta)
 [1] 0.24384466 0.02762992 0.46147411 0.57257377 0.13092734 0.01180250
 [7] 0.73116633 0.63927495 0.49887682 0.10780993 0.20213635
> reslm$coefficients
(Intercept)          X1          X2          X3          X4          X5 
 0.31205612  0.03818310  0.43924384  0.54862464  0.16904831  0.01388965 
         X6          X7          X8          X9         X10 
 0.73151975  0.64913967  0.42918589  0.13642267  0.16593451 
> res$coef
 [1] 0.31205612 0.03818310 0.43924384 0.54862464 0.16904831 0.01388965
 [7] 0.73151975 0.64913967 0.42918589 0.13642267 0.16593451

*/
